local aabb = {}

local inf = math.huge

-- Private stuff
local function abs(x)
  return x < 0 and -x or x
end

local function nearest(x, a, b)
  return abs(a - x) < abs(b - x) and a or b
end

local function liangBarsky(l,t,w,h, x1,y1,x2,y2, min, max)
  local dx, dy  = x2-x1, y2-y1
  local t0, t1  = min, max
  local p, q, r

  for side = 1,4 do
    if     side == 1 then p,q = -dx, x1 - l
    elseif side == 2 then p,q =  dx, l + w - x1
    elseif side == 3 then p,q = -dy, y1 - t
    else                  p,q =  dy, t + h - y1
    end

    if p == 0 then
      if q < 0 then return nil end
    else
      r = q / p
      if p < 0 then
        if     r > t1 then return nil
        elseif r > t0 then t0 = r
        end
      else -- p > 0
        if     r < t0 then return nil
        elseif r < t1 then t1 = r
        end
      end
    end
  end

  return t0, t1
end

local function liangBarskyCoords(l,t,w,h, x1,y1,x2,y2, min, max)
  local dx, dy = x2-x1, y2-y1
  local t0, t1 = LiangBarsky(l,t,w,h, x1,y1,x2,y2, min, max)
  if t0 then return x1 + t0*dx, y1 + t0*dy, x1 + t1*dx, y1+t1*dy end
end

local function getStaticDisplacement(l,t,w,h)
  local dx, dy  = getNearestPointInPerimeter(l,t,w,h, 0,0)
  if abs(dx) < abs(dy) then return dx, 0, 0, 0 end
  return 0, dy, 0, 0
end

-- Public stuff
--
function aabb.containsPoint(l,t,w,h, x,y)
  return not(x < l or y < t or x > l + w or y > t + h)
end
local containsPoint = aabb.containsPoint

function aabb.getNearestPointInPerimeter(l,t,w,h, x,y)
  return nearest(x, l, l+w), nearest(y, t, t+h)
end
local getNearestPointInPerimeter = aabb.getNearestPointInPerimeter

function aabb.getCenter(l,t,w,h)
  return l + w * 0.5, t + h * 0.5
end

function aabb.getMinkowskyDiff(l1,t1,w1,h1, l2,t2,w2,h2)
  return l2 - l1 - w1,
         t2 - t1 - h1,
         w1 + w2,
         h1 + h2
end
local getMinkowskyDiff = aabb.getMinkowskyDiff

function aabb.getDisplacement(l1,t1,w1,h1, l2,t2,w2,h2, dx1,dy1, dx2,dy2)

  local t
  local dx, dy = dx1 - dx2, dy1 - dy2
  local l,t,w,h = getMinkowskyDiff(l1,t1,w1,h1, l2,t2,w2,h2)

  -- if the MD contains 0,0, then the two boxes are intersecting without moving
  if containsPoint(l,t,w,h, 0,0) then
    -- if the boxes are relatively non-moving, return the minimum displacement vector
    -- to move the first box completely, and exit
    if dx == 0 and dy == 0 then return getStaticDisplacement(l,t,w,h) end

    -- if the boxes are moving, return the nearest intersection of the line containing
    -- the displacement vector with 0
    local t0, t1 = liangBarsky(l,t,w,h, 0,0,dx,dy, -inf, inf)
    t = abs(t0) < abs(t1) and t0 or t1
  else
    -- the boxes are not intersecting, so t is the intersection of MD with the
    -- relative displacement vector, if it exists
    t = liangBarsky(l,t,w,h, 0,0,dx,dy, 0,1)
  end

  if t then return dx1*t, dy1*t, dx2*t, dy2*t, t end
end


function aabb.getSegmentIntersection(l,t,w,h, x1,y1,x2,y2)
  return liangBarskyCoords(l,t,w,h, x1,y1,x2,y2, 0,1)
end

function aabb.getLineIntersection(l,t,w,h, x1,y1,x2,y2)
  return liangBarskyCoords(l,t,w,h, x1,y1,x2,y2, -inf,inf)
end


return aabb
